﻿extern alias Old;

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using Infragistics.Documents.Parsing;
using Microsoft.Win32;
using oldNS = Old.Infragistics.SyntaxParsing;

namespace Ebnf12_2to13_1Converter
{
	/// <summary>
	/// Interaction logic for MainWindow.xaml
	/// </summary>
	public partial class MainWindow : Window
	{
		#region Member Variables

		private bool hasErrors;

		#endregion // Member Variables

		#region Constructor

		public MainWindow()
		{
			InitializeComponent();
		}

		#endregion // Constructor

		#region Methods

		#region ClearLog

		private void ClearLog()
		{
			this.hasErrors = false;
			this.tbMesages.Text = string.Empty;
			this.tbMesages.Foreground = Brushes.Black;
		}

		#endregion // ClearLog

		#region ConvertComparison

		private static TerminalSymbolComparison ConvertComparison(oldNS.TerminalSymbolComparison oldComparison)
		{
			switch (oldComparison)
			{
				case oldNS.TerminalSymbolComparison.Literal:
					return TerminalSymbolComparison.Literal;

				case oldNS.TerminalSymbolComparison.LiteralIgnoreCase:
					return TerminalSymbolComparison.LiteralIgnoreCase;

				default:
				case oldNS.TerminalSymbolComparison.RegularExpression:
					return TerminalSymbolComparison.RegularExpression;
			}
		}

		#endregion // ConvertComparison

		#region ConvertFile

		private void ConvertFile(string sourcePath, string destinationPath)
		{
			this.ClearLog();
			this.LogMessage(string.Format("Converting 12.2 EBNF file from \r\n\t{0}\r\nand saving converted 13.1 EBNF file at \r\n\t{1}", sourcePath, destinationPath));
			this.LogMessage("-------------------------------------------------------");

			if (File.Exists(sourcePath) == false)
			{
				this.LogMessage("The source file does not exist.", isError: true);
				return;
			}

			try
			{
				oldNS.Grammar oldGrammar;
				Grammar newGrammar;
				this.CreateGrammarInstances(sourcePath, out oldGrammar, out newGrammar);
				if (this.hasErrors)
					return;

				this.CreateNewLexerStates(oldGrammar, newGrammar);
				if (this.hasErrors)
					return;

				this.CreateNewTerminalSymbols(oldGrammar, newGrammar);
				if (this.hasErrors)
					return;

				this.CreateNewNonTerminalSymbols(oldGrammar, newGrammar);
				if (this.hasErrors)
					return;

				this.CreateNewRuleStructures(oldGrammar, newGrammar);
				if (this.hasErrors)
					return;

				this.SetNewStartSymbol(oldGrammar, newGrammar);
				if (this.hasErrors)
					return;

				using (var outStream = new FileStream(destinationPath, FileMode.Create, FileAccess.Write))
					newGrammar.SaveEbnf(outStream);

				this.LogMessage("13.1 EBNF file created successfully");
			}
			catch (Exception ex)
			{
				this.LogMessage(ex.ToString(), isError: true);
			}
		}

		#endregion // ConvertFile

		#region CreateGrammarInstances

		private void CreateGrammarInstances(string sourcePath, out oldNS.Grammar oldGrammar, out Grammar newGrammar)
		{
			oldGrammar = null;
			newGrammar = null;

			var sourceContent = File.ReadAllText(sourcePath);
			var parser = new oldNS.Ebnf.EbnfParser();
			var result = parser.Parse(sourceContent);
			if (result.Success == false || result.Errors.Count != 0)
			{
				this.LogMessage("Cannot parse 12.2 EBNF content:", isError: true);
				foreach (var error in result.Errors)
					this.LogMessage("\t" + error.Description);

				return;
			}

			oldGrammar = parser.Grammar;
			newGrammar = new Grammar();

			if (this.cbIncludeBuiltInSymbols.IsChecked == true)
			{
				newGrammar.LexerStates.DefaultLexerState.Symbols.Add(newGrammar.WhitespaceSymbol);
				newGrammar.LexerStates.DefaultLexerState.Symbols.Add(newGrammar.NewLineSymbol);
			}

			newGrammar.Name = oldGrammar.Name;

			if (string.IsNullOrEmpty(oldGrammar.NewLineSymbolName) == false)
				newGrammar.NewLineSymbolName = oldGrammar.NewLineSymbolName;

			try
			{
				newGrammar.WhitespacePattern = oldGrammar.WhitespacePattern;
			}
			catch (ArgumentException ex)
			{
				this.LogMessage(
					string.Format("There was a problem creating the whitespace symbol in the newer format:{0}{1}",
						Environment.NewLine,
						ex.Message),
					isError: true);
				return;
			}

			if (string.IsNullOrEmpty(oldGrammar.WhitespaceSymbolName) == false)
				newGrammar.WhitespaceSymbolName = oldGrammar.WhitespaceSymbolName;
		}

		#endregion // CreateGrammarInstances

		#region CreateNewLexerStates

		private void CreateNewLexerStates(oldNS.Grammar oldGrammar, Grammar newGrammar)
		{
			foreach (var oldLexerState in oldGrammar.LexerStates)
			{
				var lexerState = newGrammar.LexerStates.Add(oldLexerState.Name);

				if (this.cbIncludeBuiltInSymbols.IsChecked == true)
				{
					lexerState.Symbols.Add(newGrammar.WhitespaceSymbol);
					lexerState.Symbols.Add(newGrammar.NewLineSymbol);
				}
			}
		}

		#endregion // CreateNewLexerStates

		#region CreateNewNonTerminalSymbols

		private void CreateNewNonTerminalSymbols(oldNS.Grammar oldGrammar, Grammar newGrammar)
		{
			foreach (var oldNonTerminalSymbol in oldGrammar.NonTerminalSymbols)
			{
				var newNonTerminalSymbol = newGrammar.NonTerminalSymbols.Add(oldNonTerminalSymbol.Name);
				newNonTerminalSymbol.HasPriority = oldNonTerminalSymbol.HasPriority;
			}
		}

		#endregion // CreateNewNonTerminalSymbols

		#region CreateNewRuleStructures

		private void CreateNewRuleStructures(oldNS.Grammar oldGrammar, Grammar newGrammar)
		{
			foreach (var oldNonTerminalSymbol in oldGrammar.NonTerminalSymbols)
				newGrammar.NonTerminalSymbols[oldNonTerminalSymbol.Name].Rule = CreateNewSyntaxRules(newGrammar, oldNonTerminalSymbol);
		}

		#endregion // CreateNewRuleStructures

		#region CreateNewSyntaxRule

		private static SyntaxRule CreateNewSyntaxRule(Grammar newGrammar, oldNS.SyntaxRuleBase oldRule)
		{
			var oldTerminalRule = oldRule as oldNS.TerminalRule;
			if (oldTerminalRule != null)
			{
				if (oldTerminalRule.TerminalSymbol.Name == "$")
					return new SymbolReferenceSyntaxRule(newGrammar.EndOfStreamSymbol);

				if (oldTerminalRule.TerminalSymbol.LanguageElement == oldNS.LanguageElement.Unrecognized)
					return new SymbolReferenceSyntaxRule(newGrammar.UnrecognizedSymbol);

				return new SymbolReferenceSyntaxRule(newGrammar.SymbolFromName(oldTerminalRule.TerminalSymbol.Name));
			}

			var oldNonTerminalRule = oldRule as oldNS.NonTerminalRule;
			if (oldNonTerminalRule != null)
				return new SymbolReferenceSyntaxRule(newGrammar.SymbolFromName(oldNonTerminalRule.NonTerminalSymbol.Name));

			var oldAlternationRule = oldRule as oldNS.AlternationRule;
			if (oldAlternationRule != null)
			{
				var newAlternationRule = new AlternationSyntaxRule();

				foreach (var oldChildRule in oldAlternationRule.Rules)
					newAlternationRule.Rules.Add(CreateNewSyntaxRule(newGrammar, oldChildRule));

				return newAlternationRule;
			}

			var oldConcatenationRule = oldRule as oldNS.ConcatenationRule;
			if (oldConcatenationRule != null)
			{
				var newConcatenationRule = new ConcatenationSyntaxRule();

				foreach (var oldChildRule in oldConcatenationRule.Rules)
					newConcatenationRule.Rules.Add(CreateNewSyntaxRule(newGrammar, oldChildRule));

				return newConcatenationRule;
			}

			var oldOptionalRule = oldRule as oldNS.OptionalRule;
			if (oldOptionalRule != null)
			{
				var newOptionalRule = new OptionalSyntaxRule();

				if (oldOptionalRule.Rules.Count == 1)
				{
					newOptionalRule.Rule = CreateNewSyntaxRule(newGrammar, oldOptionalRule.Rules[0]);
				}
				else
				{
					var newConcatenationRule = new ConcatenationSyntaxRule();

					foreach (var oldChildRule in oldOptionalRule.Rules)
						newConcatenationRule.Rules.Add(CreateNewSyntaxRule(newGrammar, oldChildRule));

					newOptionalRule.Rule = newConcatenationRule;
				}

				return newOptionalRule;
			}

			var oldRepetitionRule = oldRule as oldNS.RepetitionRule;
			if (oldRepetitionRule != null)
			{
				var newRepetitionRule = new RepetitionSyntaxRule();

				if (oldRepetitionRule.Rules.Count == 1)
				{
					newRepetitionRule.Rule = CreateNewSyntaxRule(newGrammar, oldRepetitionRule.Rules[0]);
				}
				else
				{
					var newConcatenationRule = new ConcatenationSyntaxRule();

					foreach (var oldChildRule in oldRepetitionRule.Rules)
						newConcatenationRule.Rules.Add(CreateNewSyntaxRule(newGrammar, oldChildRule));

					newRepetitionRule.Rule = newConcatenationRule;
				}

				return newRepetitionRule;
			}


			throw new InvalidOperationException("Incompatible rule type: " + oldRule.GetType().Name);
		}

		#endregion // CreateNewSyntaxRule

		#region CreateNewSyntaxRules

		private static SyntaxRule CreateNewSyntaxRules(Grammar newGrammar, oldNS.NonTerminalSymbol oldNonTerminalSymbol)
		{
			if (oldNonTerminalSymbol.Rules.Count == 1)
				return CreateNewSyntaxRule(newGrammar, oldNonTerminalSymbol.Rules[0]);

			var rootRule = new AlternationSyntaxRule();

			foreach (var oldRule in oldNonTerminalSymbol.Rules)
				rootRule.Rules.Add(CreateNewSyntaxRule(newGrammar, oldRule));

			return rootRule;
		}

		#endregion // CreateNewSyntaxRules

		#region CreateNewTerminalSymbolInLexerState

		private TerminalSymbol CreateNewTerminalSymbolInLexerState(oldNS.TerminalSymbol oldTerminalSymbol, LexerState lexerState, bool isExitSymbol)
		{
			TerminalSymbol newTerminalSymbol = null;
			try
			{
				newTerminalSymbol = lexerState.Symbols.Add(
					oldTerminalSymbol.Name,
					oldTerminalSymbol.Value,
					ConvertComparison(oldTerminalSymbol.Comparison),
					isExitSymbol);
			}
			catch (ArgumentException ex)
			{
				this.LogMessage(
					string.Format("There was a problem creating the symbol {0} in the newer format:{1}{2}",
						oldTerminalSymbol.Name,
						Environment.NewLine,
						ex.Message),
					isError: true);
				return null;
			}

			var boundaryPattern = oldTerminalSymbol.BoundaryPattern;
			if (string.IsNullOrEmpty(boundaryPattern) == false)
			{
				boundaryPattern = boundaryPattern.Replace(@"\Z", @"\z");
				newTerminalSymbol.LookaheadPattern = boundaryPattern;
			}

			newTerminalSymbol.LanguageElementName = oldTerminalSymbol.LanguageElement.ToString();

			if (IsSignificant(newTerminalSymbol) != oldTerminalSymbol.IsSignificant)
				newTerminalSymbol.IsSignificant = oldTerminalSymbol.IsSignificant;

			return newTerminalSymbol;
		}

		#endregion // CreateNewTerminalSymbolInLexerState

		#region CreateNewTerminalSymbols

		private void CreateNewTerminalSymbols(oldNS.Grammar oldGrammar, Grammar newGrammar)
		{
			foreach (var oldTerminalSymbol in oldGrammar.TerminalSymbols)
				this.CreateTerminalSymbolInNewGrammar(oldGrammar, newGrammar, oldTerminalSymbol);
		}

		#endregion // CreateNewTerminalSymbols

		#region CreateTerminalSymbolInNewGrammar

		private void CreateTerminalSymbolInNewGrammar(oldNS.Grammar oldGrammar, Grammar newGrammar, oldNS.TerminalSymbol oldTerminalSymbol)
		{
			// The unrecognized symbol is built in for 13.1 and should not be copied over.
			if (oldTerminalSymbol.LanguageElement == oldNS.LanguageElement.Unrecognized)
				return;

			TerminalSymbol newTerminalSymbol = null;
			LexerState lexerStateToEnter = null;

			foreach (var oldLexerState in oldGrammar.LexerStates)
			{
				var newLexerState = newGrammar.LexerStates[oldLexerState.Name];

				if (oldLexerState.EnterSymbol == oldTerminalSymbol)
					lexerStateToEnter = newLexerState;

				var isExitSymbol = oldLexerState.ExitSymbol == oldTerminalSymbol;
				if (isExitSymbol || oldLexerState.Symbols.Contains(oldTerminalSymbol))
				{
					if (newTerminalSymbol == null)
					{
						newTerminalSymbol = this.CreateNewTerminalSymbolInLexerState(oldTerminalSymbol, newLexerState, isExitSymbol);
						if (newTerminalSymbol == null)
							return;
					}
					else
					{
						newLexerState.Symbols.Add(newTerminalSymbol, isExitSymbol);
					}
				}
			}

			if (newTerminalSymbol == null)
			{
				newTerminalSymbol = this.CreateNewTerminalSymbolInLexerState(oldTerminalSymbol, newGrammar.LexerStates.DefaultLexerState, isExitSymbol: false);
				if (newTerminalSymbol == null)
					return;
			}

			if (lexerStateToEnter != null)
				newTerminalSymbol.LexerStateToEnter = lexerStateToEnter;
		}

		#endregion // CreateTerminalSymbolInNewGrammar

		#region IsSignificant

		private static bool IsSignificant(TerminalSymbol newTerminalSymbol)
		{
			switch (newTerminalSymbol.LanguageElement)
			{
				case LanguageElement.Whitespace:
				case LanguageElement.NewLine:
				case LanguageElement.Comment:
					return false;

				default:
					return true;
			}
		}

		#endregion // IsSignificant

		#region LogMessage

		private void LogMessage(string message, bool isError = false)
		{
			this.tbMesages.Text += string.Format("{0}{1}{2}",
				isError ? "Error: " : string.Empty,
				message,
				Environment.NewLine);

			if (isError)
			{
				this.hasErrors = true;
				this.tbMesages.Foreground = Brushes.Red;
			}
		}

		#endregion // LogMessage

		#region SetNewStartSymbol

		private void SetNewStartSymbol(oldNS.Grammar oldGrammar, Grammar newGrammar)
		{
			newGrammar.StartSymbol = newGrammar.NonTerminalSymbols[oldGrammar.StartSymbol.Name];

			// Fix-up the start symbol if it had an EOS token at the end of it but will now be invalid.
			var startSymbol = newGrammar.StartSymbol;
			var concatenationRule = startSymbol.Rule as ConcatenationSyntaxRule;
			if (concatenationRule == null)
				return;

			var lastConcatenatedRule = concatenationRule.Rules[concatenationRule.Rules.Count - 1] as SymbolReferenceSyntaxRule;
			if (lastConcatenatedRule == null || lastConcatenatedRule.Symbol != newGrammar.EndOfStreamSymbol)
				return;

			// If the symbol does have an EOS token as the last symbol, but it is well formed, we can allow it.
			if (concatenationRule.Rules.Count == 2 &&
				concatenationRule.Rules[0].Type == SyntaxRuleType.SymbolReference &&
				((SymbolReferenceSyntaxRule)concatenationRule.Rules[0]).Symbol is NonTerminalSymbol)
			{
				return;
			}

			// Otherwise, remove the EOS token. If that leaves multiple rules in the concatenation, the concatenation is still valid and we can return.
			concatenationRule.Rules.RemoveAt(concatenationRule.Rules.Count - 1);
			if (concatenationRule.Rules.Count > 1)
				return;

			// Otherwise, if only 1 child is left in the concatenation, the concatenation is invalid and the child must become the root of the rule tree.
			var childRule = concatenationRule.Rules[0];
			concatenationRule.Rules.RemoveAt(0);
			startSymbol.Rule = childRule;
		}

		#endregion // SetNewStartSymbol

		#endregion // Methods

		#region Event Handlers

		private void btConvert_Click(object sender, RoutedEventArgs e)
		{
			this.ConvertFile(this.tbSourcePath.Text, this.tbDestinationPath.Text);
		}

		private void btCopyToClipboard_Click(object sender, RoutedEventArgs e)
		{
			Clipboard.SetText(this.tbMesages.Text);
		}

		private void btDestinationBrowse_Click(object sender, RoutedEventArgs e)
		{
			var sfd = new SaveFileDialog();
			sfd.Title = "Please select the destination 13.1 EBNF file path";
			sfd.DefaultExt = ".ebnf";
			sfd.Filter = "EBNF documents|*.ebnf|All Documents|*.*";

			if (sfd.ShowDialog() == true)
				this.tbDestinationPath.Text = sfd.FileName;
		}

		private void btSourceBrowse_Click(object sender, RoutedEventArgs e)
		{
			var ofd = new OpenFileDialog();
			ofd.Title = "Please select the source 12.2 EBNF file path";
			ofd.Multiselect = false;
			ofd.DefaultExt = ".ebnf";
			ofd.Filter = "EBNF documents|*.ebnf|All Documents|*.*";

			if (ofd.ShowDialog() == true)
				this.tbSourcePath.Text = ofd.FileName;
		}

		private void pathTextBoxes_TextChanged(object sender, TextChangedEventArgs e)
		{
			this.btConvert.IsEnabled =
				String.IsNullOrEmpty(this.tbDestinationPath.Text) == false &&
				String.IsNullOrEmpty(this.tbSourcePath.Text) == false;
		}

		#endregion // Event Handlers
	}
}
